import pandas as pd
import logging
import tqdm

from fhir2dataset.graphquery import GraphQuery
from fhir2dataset.fhirrules_getter import FHIRRules
from fhir2dataset.api_caller import ApiGetter
from fhir2dataset.url_builder import URLBuilder
from fhir2dataset.graph_tools import join_path

logger = logging.getLogger(__name__)


class Query:
    """Query Executor

    Description:
        This module purpose is to perform SQL-type queries whose information is
        filled in configuration files of the form:
        ```
        {
            "from": {
                "alias n°1": "Resource type 1",
                "alias n°2": "Resource type 2",
                "alias n°3": "Resource type 3",
                ...
            },
            "select": {
                "alias n°1": [
                    "searchparam/fhirpath of attribute a of resource type 1",
                    "searchparam/fhirpath of attribute b of resource type 1",
                    "searchparam/fhirpath of attribute c of resource type 1"
                ],
                "alias n°2": [
                    "searchparam of attribute a of resource type 2",
                    ...
                ],
                ...
            },
            "join": {
                "inner": {
                    "alias n°1": {
                        "searchparam of attribute d, which is of type Reference, of resource type 1": "alias n°2"
                    },
                    ...
                },
                "child": {
                    "alias n°2": {
                        "searchparam of attribute b, which is of type Reference, of resource type 2": "alias n°3"
                    },
                    ...
                },
                "parent": {
                    ...
                }
            },
            "where": {
                "alias n°2": {
                    "searchparam of attribute c of resource type 2": "value 1",
                    "searchparam of attribute d of resource type 2": "value 2"
                },
                "alias n°3": {
                    "searchparam of attribute a of resource type 2": "value 3",
                    "searchparam of attribute b of resource type 2": "value 4"
                },
                ...
            }
        }
        ```
        for the next associated SQL query:
        ```
        SELECT (alias n°1).a, (alias n°1).b, (alias n°1).c, (alias n°2).a FROM (Resource type 1) as (alias n°1)
        INNER JOIN (Resource type 2) as (alias n°2)
        ON (alias n°1).d = (alias n°2)
        INNER JOIN (Resource type 3) as (alias n°3)
        ON (alias n°2).b = (alias n°3) WHERE (alias n°2).c = "value 1"
        AND (alias n°2).d = "value 2"
        AND (alias n°3).a = "value 3"
        AND (alias n°3).b = "value 4"
    ```

    Attributes:
        config {dict} -- dictionary storing the initial request
        graph_query {type(GraphQuery)} -- instance of a GraphQuery object that gives a graphical representation of the query
        dataframes {dict} -- dictionary storing for each alias the resources requested on the api in tabular format
        main_dataframe {DataFrame} -- pandas dataframe storing the final result table

    Example:
        import fhir2dataset
        import logging

        logging.basicConfig(level=logging.INFO)
        query = Query(fhir_api_url=fhir_api_url, fhir_rules=fhir_rules, token=token)
        query.from_config(config)
        query.execute()
        df = query.main_dataframe
    """  # noqa

    def __init__(
        self,
        fhir_api_url: str = None,
        fhir_rules: type(FHIRRules) = None,
        token: str = None,
    ):
        """Requestor's initialisation

        Arguments:
            fhir_api_url {str} -- The Service Base URL (e.g. http://hapi.fhir.org/baseR4/)

        Keyword Arguments:
            fhir_rules {type(FHIRRules)} -- an instance of a FHIRRules-type object. If the instance is not filled a default version will be used. (default: {None})
            token {str} -- bearer token authentication if necessary (default: {None})
        """  # noqa
        self.fhir_api_url = fhir_api_url or "http://hapi.fhir.org/baseR4/"
        if not fhir_rules:
            fhir_rules = FHIRRules(fhir_api_url=self.fhir_api_url)
        self.fhir_rules = fhir_rules
        self.token = token

        self.config = None
        self.graph_query = None
        self.dataframes = {}
        self.main_dataframe = None

    def from_config(self, config: dict):
        """Executes the query from a dictionary in the format of a configuration file

        Arguments:
            config {dict} -- dictionary in the format of a configuration file
        """
        self.config = {
            "from_dict": config.get("from", None),
            "select_dict": config.get("select", None),
            "where_dict": config.get("where", None),
            "join_dict": config.get("join", None),
        }
        return self

    def execute(self, debug: bool = False):
        """Executes the complete query

        1. constructs a GraphQuery object to store the query as a graph
        2. builds the url of the requests to send to the API
        3. retrieves the answers from the api and puts them as tables in the dataframes attribute
        4. executes joins to return the result table in the main_dataframe attribute

        Keyword Arguments::
            debug {bool} -- if debug is true then the columns needed for internal processing are kept in the final dataframe. Otherwise only the columns of the select are kept in the final dataframe. (default: {False})
        """  # noqa
        self.graph_query = GraphQuery(fhir_api_url=self.fhir_api_url, fhir_rules=self.fhir_rules)
        self.graph_query.execute(**self.config)
        with tqdm.tqdm(total=1000) as pbar:
            time_frac = round(1000 / len(self.graph_query.resources_alias_info))
            for resource_alias in self.graph_query.resources_alias_info.keys():
                resource_alias_info = self.graph_query.resources_alias_info[resource_alias]
                elements = resource_alias_info.elements
                url_builder = URLBuilder(
                    fhir_api_url=self.fhir_api_url,
                    graph_query=self.graph_query,
                    main_resource_alias=resource_alias,
                )
                url = url_builder.compute()
                call = ApiGetter(
                    url=url, elements=elements, token=self.token, pbar=pbar, time_frac=time_frac
                )
                call.get_all()
                self.dataframes[resource_alias] = call.df
        self._clean_columns()
        for resource_alias, dataframe in self.dataframes.items():
            logger.debug(f"{resource_alias} dataframe builded head - \n{dataframe.to_string()}")
        # We check if there is more than 1 alias of resource
        if len(self.dataframes) > 1:
            self.main_dataframe = self._join()
        else:
            self.main_dataframe = list(self.dataframes.values())[0]

        logger.debug(
            f"Main dataframe builded head before columns selection-"
            f"\n{self.main_dataframe.to_string()}"
        )
        if not debug:
            self._select_columns()
            self._remove_lists()

    def _select_columns(self):
        """Clean the final dataframe to keep only the columns of the select"""
        final_columns = []
        for resource_alias, resource_alias_info in self.graph_query.resources_alias_info.items():
            for element in resource_alias_info.elements.get_subset_elements(goal="select"):
                final_columns.append(f"{resource_alias}:{element.col_name}")
        self.main_dataframe = self.main_dataframe[final_columns]

    def _remove_lists(self):
        """Remove lists from columns with only single elements"""

        def unlist(x):
            """Auxiliary function to squeeze lists with one element"""
            if isinstance(x, list) and len(x) == 1:
                return unlist(x[0])
            else:
                return x

        if len(self.main_dataframe) > 0:
            df = self.main_dataframe.copy()
            for column in df.columns:
                df[column] = df[column].apply(unlist)
            self.main_dataframe = df

    def _join(self) -> pd.DataFrame:
        """Execute the joins one after the other in the order specified by the
        join_path function.

        Returns:
            pd.DataFrame -- dataframe containing all joined resources
        """
        list_join = join_path(self.graph_query.resources_alias_graph)
        main_alias_join = list_join[0][0]
        main_df = self.dataframes[main_alias_join]
        for alias_1, alias_2 in list_join:
            df_1 = main_df
            df_2 = self.dataframes[alias_2]
            main_df = self._join_2_df(alias_1, alias_2, df_1, df_2)
        return main_df

    def _group_lines(self, df, col_name):
        logger.debug(f"dataframe before being grouped \n{df.to_string()}")
        if not df.empty:
            cols_group = [col_name]
            if cols_group:
                cols = df.columns.to_list()
                cols_list = [col_name for col_name in cols if col_name not in cols_group]
                dict_cols_list = {col: self._concatenate for col in cols_list}
                df = df.groupby(cols_group).agg(dict_cols_list)
                df.reset_index(inplace=True)
            logger.debug(f"dataframe after being grouped on {cols_group}:\n{df.to_string()}")
        return df

    def _concatenate(self, column):
        result = []
        for list_cell in column:
            if isinstance(list_cell, list):
                result.extend(list_cell)
            else:
                result.append(list_cell)
        return result

    def _join_2_df(
        self, alias_1: str, alias_2: str, df_1: pd.DataFrame, df_2: pd.DataFrame
    ) -> pd.DataFrame:
        """Executes the join between two dataframes

        The join key is the id of the child resource.
        This id is contained in :
            * in the column child_alias:id of the child resource table named child_alias (e.g. parent:id for a parent dataframe)
            * in the alias_parent:searchparam_parent column of the parent resource table named alias_parent (e.g. condition:subject.reference for a condition dataframe)

        The function is in charge of finding out who is the mother resource and who is the daughter resource.

        Arguments:
            alias_1 {str} -- df_1 alias
            alias_2 {str} -- df_2 alias
            df_1 {pd.DataFrame} -- dataframe containing the elements of a resource
            df_2 {pd.DataFrame} -- dataframe containing the elements of a resource
        Returns:
            pd.DataFrame -- dataframe containing the elements of the 2 resources according to an inner join
        """  # noqa
        edge_info = self.graph_query.resources_alias_graph.edges[alias_1, alias_2]["info"]
        alias_parent = edge_info.parent
        alias_child = edge_info.child

        how = edge_info.join_how
        searchparam_parent = edge_info.searchparam_parent

        parent_on = f"{alias_parent}:join_{searchparam_parent}"
        child_on = f"{alias_child}:from_id"

        if how == "child":
            how = "right"
        elif how == "parent":
            how = "left"
        elif how != "inner":
            how = "inner"

        if alias_1 == alias_parent:
            df_merged_inner = pd.merge(
                left=df_1, right=df_2, left_on=parent_on, right_on=child_on, how=how
            )
        else:
            df_merged_inner = pd.merge(
                left=df_2, right=df_1, left_on=parent_on, right_on=child_on, how=how
            )
        return df_merged_inner

    def _clean_columns(self):
        """performs preprocessing on all dataframes harvested in the dataframe attribute:
        * adds the resource type in front of an element id so that the resource id matches the references of its parent resource references
        * adds the table alias as a prefix to each column name
        """  # noqa
        for resource_alias, df in self.dataframes.items():
            resource_type = self.graph_query.resources_alias_info[resource_alias].resource_type

            df = df.pipe(_add_resource_type_to_id, resource_type=resource_type)

            self.dataframes[resource_alias] = df.add_prefix(f"{resource_alias}:")


def _add_resource_type_to_id(df, resource_type: str):
    # add assert to check that there is only one id in list
    df["from_id"] = f"{resource_type}/" + df["from_id"]
    return df



"""set of functions allowing to use the javascript coded library on the repository https://github.com/HL7/fhirpath.js
"""  # noqa
import os
import json
import logging
from subprocess import Popen, PIPE
from typing import List


logger = logging.getLogger(__name__)

wrapper = """
(function run(globals) {
    try {
        let result = (%(func)s).apply(globals, %(args)s);
        if ((typeof result) == 'string') {
            result = JSON.stringify(result);
        }
        console.log(
            %(result_keyword)s +
            JSON.stringify({"result": result}) +
            %(result_keyword)s
            );
    } catch (e) {
        console.log(
            %(result_keyword)s +
            JSON.stringify({error: e.message}) +
            %(result_keyword)s
            );
    }
})(%(globals)s);
"""


def execute(code: str, args: list = None, g: dict = None):
    """Function to execute code written in javascript

    Args:
        code (str): javascript code
        args (list, optional): the arguments that the js code takes as input. Defaults to None.
        g (dict, optional): the 'this' object in the code. Defaults to None.

    Returns:
        dict or list: result of the js script
    """
    if args is None:
        args = []

    if g is None:
        g = {}

    assert isinstance(code, str)
    assert isinstance(g, dict)
    assert code.strip(" ").strip("\t").startswith("function"), "Code must be function"

    # Here, the Popen function allows to execute the javascript script as in a terminal with the
    # command 'node'. As in a terminal, the result is returned in console between the keywords
    # defined by the variable VV. The python script, is in charge of retrieving the result between
    # these keywords.
    path = os.path.join(os.path.dirname(__file__), "metadata")

    prc = Popen(
        "node", shell=False, stderr=PIPE, stdout=PIPE, stdin=PIPE, cwd=path, encoding="utf-8"
    )

    result_keyword = "--FHIRPATH--"

    c = wrapper % {
        "func": code,
        "globals": json.dumps(g),
        "args": json.dumps(args),
        "result_keyword": f'"{result_keyword}"',
    }
    outs, errs = prc.communicate(input=c)
    if isinstance(outs, str):
        outs = outs.split(result_keyword)[1]
        outs = json.loads(outs)
    if "result" in outs:
        outs = outs["result"]
        return outs
    elif "error" in outs:
        raise Exception((outs.get("error"), errs))
    else:
        raise Exception(errs)


def parse_fhirpath(fhirpath: str):
    result = execute(
        """function test(fhirpath){
        const fhirpath_module = require("fhirpath");
        return JSON.stringify(fhirpath_module.parse(fhirpath))
    }
    """,
        args=[fhirpath],
    )
    return result


def fhirpath_processus_tree(forest_dict, resource):
    result = execute(
        """function test(args){
            const graph = require('./forest')
            forest_dict = args[0]
            resource = args[1]
            let result = []
            for (const [root_id, tree_raw] of Object.entries(forest_dict)) {
                nodes_dict_raw = tree_raw["nodes_dict"]
                edges_array_raw = tree_raw["edges_array"]
                graph.compute_fhirpaths(nodes_dict_raw, edges_array_raw, root_id, resource, result)
            }
            return result
        }
        """,
        args=[[forest_dict, resource]],
    )
    return result


def multiple_search_dict(resources: list, elements: dict) -> List[dict]:
    """Returns the updated element instance on each element in the Resources list. These updated
    instances are stored in a list which is the element returned by the function.
    The update consists for each element of the instance elements in the application of the fhirpath
    (element.fhirpath) on an instance of the Resources list and the storage of the response in element.value.

    Args:
        resources (list): list composed of several instances of resources
        elements (dict): an instance of the Elements object in dictionary format

    Returns:
        list: elements object list copied and updated on each resource composing the resources list
    """  # noqa
    result = execute(
        """function test(resources) {
                const fhirpath = require("fhirpath");
                const fhirpath_r4_model = require("fhirpath/fhir-context/r4");

                return resources.map((resource) => {
                    var fhirpath_result = [];
                    for (index = 0; index < this.elements.length; index++) {
                        try {
                            this.elements[index].value = fhirpath.evaluate(
                                resource,
                                this.elements[index].fhirpath,
                                null,
                                fhirpath_r4_model
                            );
                        } catch (e) {
                            if (e.message.includes("TypeExpression")) {
                                this.elements[index].value = [
                                    "the fhirpath could not be evaluated by the library",
                                ];
                            } else {
                                throw e;
                            }
                        }
                    };
                    return JSON.parse(JSON.stringify(this))
                });
            }
    """,
        args=[resources],
        g=elements,
    )
    return result



from fhir2dataset.query import Query  # noqa
from fhir2dataset.fhirrules_getter import FHIRRules  # noqa
from fhir2dataset.parser import Parser  # noqa

import re


def sql(sql_query, fhir_api_url=None, token=None):
    config = Parser().from_sql(sql_query)
    query = Query(fhir_api_url=fhir_api_url, token=token).from_config(config)
    query.execute()
    df = query.main_dataframe
    df = df.reset_index(drop=True)
    # rename the columns to match the sql syntax
    # patient:Patient.name.given -> patient.name.given
    df = df.rename(lambda x: re.sub("\:\w+\.", ".", x), axis="columns")  # noqa
    return df.reset_index(drop=True)



"""
    Module containing functions useful for the analysis and exploitation of graphs
"""

import networkx as nx
import logging


logger = logging.getLogger(__name__)


def join_path(graph: nx.Graph) -> list:
    """transforms the query graph into an Eulerian graph in order to be able to find an Eulerian path in it.

    An Eulerian path is a trail in a finite graph that visits every edge exactly once (allowing for revisiting vertices).
    Since the initial graph is not necessarily an Eulerian graph, the Eulerian path is reprocessed so that each join is made only once.

    Arguments:
        graph {nx.Graph} -- instance of GraphQuery

    Returns:
        list -- List of tuples indicating the successive joints to be made
    """  # noqa
    euler_graph = nx.eulerize(graph)
    euler_path = list(nx.eulerian_path(euler_graph))
    path = clean_euler_path(euler_path)
    return path


def clean_euler_path(eulerian_path: list) -> list:
    """Cleans a Eulerian path so that each edge (not directed) appears only once in the list. If a edge appears more than once, only the first occurrence is kept.

    Arguments:
        eulerian_path {list} -- Eulerian path

    Returns:
        list -- cleaned Eulerian path
    """  # noqa
    path = []
    for edge in eulerian_path:
        if edge not in path and edge[::-1] not in path:
            path.append(edge)
    return path



import logging
import random
from typing import Type

from collections import defaultdict
from posixpath import join as urljoin
from fhir2dataset.graphquery import GraphQuery
from fhir2dataset.data_class import SearchParameter


logger = logging.getLogger(__name__)


class URLBuilder:
    """class that allows to build the url of a query that will be made for an alias of a FHIR resource to a FHIR api

    Attributes:
        fhir_api_url: The Service Base URL (e.g. http://hapi.fhir.org/baseR4/)
        main_resource_alias: alias given to a set of fhir resources of a certain type which are the subject of the api query
        search_query_url: url which will make it possible to recover the resources of the type of main_alias respecting as well as possible the conditions where on itself and on its neighbors.
    """  # noqa

    def __init__(
        self, fhir_api_url: str, graph_query: type(GraphQuery), main_resource_alias: str
    ) -> None:
        """
        Arguments:
            fhir_api_url (str} -- The Service Base URL (e.g. http://hapi.fhir.org/baseR4/)
            graph_query {type(GraphQuery)} -- instance of a GraphQuery object that gives a graphical representation of the global query
            main_resource_alias {str} -- alias given to a set of fhir resources of a certain type
        """  # noqa
        self.fhir_api_url = fhir_api_url
        self.main_resource_alias = main_resource_alias
        self._graph_query = graph_query

        self._params = defaultdict(list)

    def compute(self):
        self._update_url_params()
        self.search_query_url = self._compute_url()
        return self.search_query_url

    def _compute_url(self) -> str:
        """Generates the url which will make it possible to recover the resources of the type of main_alias respecting as well as possible the conditions where on itself and on its neighbors.

        Returns:
            str -- corresponding API request url
        """  # noqa
        params = f"{self._graph_query.resources_alias_info[self.main_resource_alias].resource_type}?"  # noqa

        for key, values in self._params.items():
            if not isinstance(values, list):
                values = [values]
            for value in values:
                params += f"{key}={value}&"

        # Rm last '&'
        params = params[:-1]

        search_query_url = urljoin(self.fhir_api_url, params)
        logger.info(f"the computed url is {search_query_url}")
        return search_query_url

    def _update_url_params(self):
        """retrieves the portions of the url that specify search parameters

        The FHIR API makes union when "where conditions" are requested for neighbouring resources 
        Only one "where condition" on every neighbouring resource is taken into account
        """  # noqa
        for resource_alias in self._graph_query.resources_alias_graph.neighbors(
            self.main_resource_alias
        ):
            edge_info = self._graph_query.resources_alias_graph.edges[
                self.main_resource_alias, resource_alias
            ]["info"]

            resource_alias_info = self._graph_query.resources_alias_info[resource_alias]
            elements = resource_alias_info.elements.get_subset_elements(goal="where")

            # check if there are "where conditions" on resource_alias
            if elements:
                # select a random "where condition" on resource_alias
                element = random.choice(list(elements))
                search_param = element.search_parameter
                searchparam_prefixe = edge_info.searchparam_prefix[self.main_resource_alias]

                self._update_params_dict(search_param, searchparam_prefixe=searchparam_prefixe)

                logger.debug(f"the part of the url for the params is: {self._params}")

        # "where condition" on the resource itself
        resource_alias_info = self._graph_query.resources_alias_info[self.main_resource_alias]
        elements = resource_alias_info.elements.get_subset_elements(goal="where")

        for element in elements:
            self._update_params_dict(element.search_parameter)

        logger.debug(f"the part of the url for the params is: {self._params}")

    def _update_params_dict(
        self, search_param: Type[SearchParameter], searchparam_prefixe: str = ""
    ):
        key = f"{searchparam_prefixe}{search_param.code}"
        value = f"{search_param.prefix or ''}{search_param.value}"

        if key not in self._params:
            self._params[key] = []
        self._params[key].append(value)



import networkx as nx
import logging
from typing import Type
from pprint import pformat
from collections import defaultdict

from fhir2dataset.fhirrules_getter import FHIRRules
from fhir2dataset.visualization_tools import custom_repr

from fhir2dataset.data_class import SearchParameter, ResourceAliasInfo, Element, Elements, EdgeInfo

logger = logging.getLogger(__name__)


class GraphQuery:
    """Class for storing query information in the form of a graph.

    Attributes:
        resources_alias_graph {nx.Graph} -- The nodes correspond to the aliases involved in the query (filled in the "from") and stops the reference link between 2 aliases (filled in the "join").
        resources_alias_info {dict} -- Dictionary storing information about each alias:
                                        * the type of the associated resource
                                        * the elements that must be retrieved from the json of a resource
                                        * a boolean indicating whether to return the number of instances of the resource that meets all these criteria or not
        fhir_rules {Type(FHIRRules)} -- an instance of an FHIRRules object which contains information specific to the FHIR standard and the API used (for example the fhirpaths associated with the search param of a resource).
    """  # noqa

    def __init__(self, fhir_api_url: str, fhir_rules: type(FHIRRules) = None) -> None:
        """Instantiate the class and create the query object

        Arguments:
            fhir_api_url {str} -- the Service Base URL (e.g. http://hapi.fhir.org/baseR4/)

        Keyword Arguments:
            fhir_rules {type(FHIRRules)} -- an instance of a FHIRRules-type object. If the instance is not filled a default version will be used. (default: {None})
        """  # noqa
        self.fhir_api_url = fhir_api_url
        self.fhir_rules = fhir_rules or FHIRRules(fhir_api_url=self.fhir_api_url)

        # to represent the relationships (references) between resources
        self.resources_alias_graph = nx.Graph()
        self.resources_alias_info = defaultdict(Type[ResourceAliasInfo])

    def execute(
        self,
        select_dict: dict,
        from_dict: dict,
        join_dict: dict = None,
        where_dict: dict = None,
        default_element_concat_type: str = "cell",
    ):
        """Populates the attributes resources_alias_graph and resources_alias_info according to the information filled in

        Arguments:
            select_dict {dict} -- dictionary containing the elements to be selected from the different resources
            from_dict {dict} -- dictionary containing all requested resources

        Keyword Arguments:
            join_dict {dict} -- dictionary containing the inner join rules between resources (default: {None})
            where_dict {dict} -- dictionary containing the (cumulative) conditions to be met by the resources (default: {None})
            default_element_concat_type {str} -- indicates how multiple occurrences of elements should be concatenated (cell: all in one cell, row: split in several rows, column: split in several columns) (default: {"cell"})
        """  # noqa
        self._from(**from_dict)
        if join_dict:
            self._join(**join_dict)
        if where_dict:
            self._where(**where_dict)
        self._select(**select_dict)

        for resource_alias in self.resources_alias_info.keys():
            self.resources_alias_info[resource_alias].elements.compute_forest_fhirpaths()

        logger.info(f"The nodes are:{self.resources_alias_graph.nodes()}")
        logger.info("The edges are:")
        logger.info(pformat(list(self.resources_alias_graph.edges(data=True))))
        logger.info("The information gathered for each node is:")
        logger.info(pformat(self.resources_alias_info))

    def from_config(self, config: dict):
        """Populates the attributes resources_alias_graph and resources_alias_info according to the information given in the configuration file

        Arguments:
            config {dict} -- dictionary in the format of a configuration file
        """  # noqa
        self.execute(
            from_dict=config.get("from"),
            select_dict=config.get("select"),
            where_dict=config.get("where"),
            join_dict=config.get("join"),
        )

    def _from(self, **resource_type_alias):
        """Initializes the graph nodes contained in resources_alias_graph and the dictionary of resources_alias_info information of the aliases listed in resource_type_alias

        Keyword Arguments:
            **resource_type_alias: the key corresponds to the alias and the value to the type of the resource.
        """  # noqa
        for (resource_alias, resource_type) in resource_type_alias.items():
            elements = Elements(
                [
                    Element(
                        goal="additional_resource",
                        col_name="from_id",
                        fhirpath="id",
                        concat_type="row",
                    )
                ]
            )

            self.resources_alias_graph.add_node(resource_alias)

            self.resources_alias_info[resource_alias] = ResourceAliasInfo(
                alias=resource_alias, resource_type=resource_type, elements=elements
            )

    def _join(self, **join_as):
        """Builds the reference links between the aliases involved in the query
        1. fills in the elements in attribute resources_alias_info to be retrieved from the json resource file to be able to make the joins
        2. creates the edges between the nodes of the relevant aliases in the resources_alias_graph attribute.

        Keyword Arguments:
            **join_as: the key is an alias of a parent resource and the value is a dictionary containing in the key the search parameter leading to the reference, and in the value the alias of the child resource.
        """  # noqa
        # TODO : review naming : element not very precise
        for join_how, relationships_dict in join_as.items():
            join_how = join_how.lower()
            assert join_how in ["inner", "child", "parent", "one"], "Precise how to join"
            for (alias_parent, searchparam_dict) in relationships_dict.items():
                type_parent = self.resources_alias_info[alias_parent].resource_type
                for (searchparam_parent, alias_child) in searchparam_dict.items():

                    # Update element to have in table
                    fhirpath_searchparam = self.fhir_rules.searchparam_to_fhirpath(
                        resource_type=type_parent, search_param=searchparam_parent
                    )
                    if " | " in fhirpath_searchparam:
                        fhirpath_ref = f"(({fhirpath_searchparam}).reference)"
                    else:
                        fhirpath_ref = f"({fhirpath_searchparam}.reference)"

                    self.resources_alias_info[alias_parent].elements.append(
                        Element(
                            goal="join",
                            col_name=f"join_{searchparam_parent}",
                            fhirpath=fhirpath_ref,
                            concat_type="row",
                        )
                    )

                    # Udpade Graph
                    type_child = self.resources_alias_info[alias_child].resource_type

                    searchparam_parent_to_child = f"{searchparam_parent}:{type_child}."
                    # include = f"{type_parent}:{searchparam_parent}:" f"{type_child}"

                    searchparam_child_to_parent = f"_has:{type_parent}:{searchparam_parent}:"
                    # revinclude = f"{type_parent}:{searchparam_parent}:" f"{type_child}"

                    searchparam_prefix = {
                        alias_parent: searchparam_parent_to_child,
                        alias_child: searchparam_child_to_parent,
                    }

                    edge_info = EdgeInfo(
                        parent=alias_parent,
                        child=alias_child,
                        searchparam_parent=searchparam_parent,
                        join_how=join_how,
                        searchparam_prefix=searchparam_prefix,
                    )

                    self.resources_alias_graph.add_edge(alias_parent, alias_child, info=edge_info)

    def _where(self, **wheres):
        """updates the resources_alias_info attribute with the conditions that each alias must meet

        Keyword Arguments:
            **wheres: the key is an alias, the value is a dictionary containing itself keys which are searchparams and whose values must be respected by the associated search params.
        """  # noqa
        for resource_alias, conditions in wheres.items():
            for search_param, list_value_full in conditions.items():
                fhirpath = self._check_searchparam_or_fhirpath(resource_alias, search_param)
                if not isinstance(list_value_full, list):
                    list_value_full = [list_value_full]

                for value_full in list_value_full:

                    # handles the case where a prefix has been specified
                    # value_full={"ge": "1970"}
                    if type(value_full) is dict:
                        for k, v in value_full.items():
                            prefix = k
                            value = v
                    else:
                        prefix = None
                        value = value_full

                    search_param_obj = SearchParameter(
                        code=search_param, prefix=prefix, value=value
                    )
                    self.resources_alias_info[resource_alias].elements.append(
                        Element(
                            goal="where",
                            col_name=f"where_{search_param_obj.code}",
                            fhirpath=fhirpath,
                            search_parameter=search_param_obj,
                        )
                    )

    def _select(self, **selects):
        """updates the resources_alias_info attribute with the elements that must be retrieved for each alias

        Keyword Arguments:
            **selects: the key is an alias, the value is a list containing fhirpath to leaf elements
        """  # noqa
        for resource_alias, col_names in selects.items():
            for col_name in col_names:
                fhirpath = self._check_searchparam_or_fhirpath(resource_alias, col_name)
                self.resources_alias_info[resource_alias].elements.append(
                    Element(goal="select", col_name=col_name, fhirpath=fhirpath)
                )

    def _check_searchparam_or_fhirpath(self, resource_alias: str, searchparam_or_fhirpath: str):
        """transforms searchparam_or_fhirpath into its fhirpath if it's a searchparam, otherwise it returns the argument as it was entered.

        Args:
            resource_alias (str): alias associated with a resource
            searchparam_or_fhirpath (str): string of characters that can correspond to a searchparam or a fhirpath

        Returns:
            str: string of characters corresponding to a fhirpath
        """  # noqa
        resource_type = self.resources_alias_info[resource_alias].resource_type
        searchparam_to_element = self.fhir_rules.searchparam_to_fhirpath(
            resource_type=resource_type,
            search_param=searchparam_or_fhirpath,
        )
        if searchparam_to_element:
            element = searchparam_to_element
        else:
            element = searchparam_or_fhirpath
        return element

    def draw_relations(self):
        """draws the resources_alias_graph attribute"""
        import matplotlib.pyplot as plt

        edge_labels = {}
        for i in self.resources_alias_graph.edges(data=True):
            edge_infos = custom_repr(i[2]["info"].__repr__())
            edge_labels[i[0:2]] = edge_infos

        plt.figure(figsize=(15, 15))
        layout = nx.spring_layout(self.resources_alias_graph)
        nx.draw_networkx(self.resources_alias_graph, pos=layout)
        nx.draw_networkx_labels(self.resources_alias_graph, pos=layout)
        nx.draw_networkx_edge_labels(
            self.resources_alias_graph,
            pos=layout,
            edge_labels=edge_labels,
            font_size=10,
            rotate=False,
            horizontalalignment="left",
        )
        plt.show()



OPEN_CHAR = ["[", "{", "("]
CLOSE_CHAR = ["]", "}", ")"]
SAFE_CHAR = ["'", '"']
DEFAULT_BREAKLINE_CHAR = "\n"
DEFAULT_INDENT_CHAR = " "


def custom_repr(
    string: str,
    indent_width: int = 4,
    breakline_char: str = DEFAULT_BREAKLINE_CHAR,
    indent_char: str = DEFAULT_INDENT_CHAR,
):
    nested_int = 0
    new_string = ""
    safe_zone = False
    len_breakline_char = len(breakline_char)
    for character in string:
        if character in SAFE_CHAR:
            safe_zone = not safe_zone
            new_string += character
        elif safe_zone:
            new_string += character
        elif character in OPEN_CHAR:
            nested_int += 1
            new_string += character + breakline_char + nested_int * indent_width * indent_char
        elif character == ",":
            if new_string[-len_breakline_char:] == breakline_char:
                new_string = new_string[:-len_breakline_char]
            new_string += "," + breakline_char + (nested_int * (indent_width) - 1) * indent_char
        elif character in CLOSE_CHAR:
            nested_int -= 1
            if new_string[-len_breakline_char:] == breakline_char:
                new_string = new_string[:-len_breakline_char]
            new_string += (
                breakline_char
                + nested_int * indent_width * indent_char
                + character
                + breakline_char
            )
        else:
            new_string += character
    return new_string



import re
from collections import defaultdict
from collections.abc import Iterable

from typing import List

PREFIX = ["eq", "ne", "gt", "lt", "ge", "le", "sa", "eb", "ap"]


def create_mask(
    separators,
    keep_separators=False,
    case_insensitive: bool = True,
    can_start_sentence: bool = False,
    optional_space_before: bool = False,
    optional_space_after: bool = False,
    optional_spaces: bool = None,
):
    if optional_spaces is not None:
        optional_space_before = optional_spaces
        optional_space_after = optional_spaces

    assert not (
        can_start_sentence and optional_space_before
    ), "Can't set to true both can_start_sentence and optional_space_before optional arguments"
    if isinstance(separators, str):
        mask = separators
    elif isinstance(separators, Iterable):
        mask = "|".join(separators)
    else:
        raise ValueError

    if keep_separators:
        mask = fr"({mask})"

    if can_start_sentence:
        mask = fr"(?:\s|^){mask}"
    elif optional_space_before:
        mask = fr"\s?{mask}"
    else:
        mask = fr"\s{mask}"

    if optional_space_after:
        mask = fr"{mask}\s?"
    else:
        mask = fr"{mask}\s"

    if case_insensitive:
        mask = fr"{mask}(?i)"

    return mask


class Parser:
    def __init__(self):
        self.CLAUSES = {
            "FROM": self.__from_parser,
            "INNER JOIN": self.__inner_join_parser,
            "CHILD JOIN": self.__child_join_parser,
            "PARENT JOIN": self.__parent_join_parser,
            "SELECT": self.__select_parser,
            "WHERE": self.__where_parser,
            "ORDER BY": self.__order_by_parser,
            "GROUP BY": self.__group_by_parser,
            "UNION": self.__union_parser,
            "LIMIT": self.__limit_parser,
        }

        self.__select = defaultdict(list)
        self.__from = defaultdict(str)
        self.__inner_join = defaultdict(dict)
        self.__child_join = defaultdict(dict)
        self.__parent_join = defaultdict(dict)
        self.__where = defaultdict(dict)

    def from_sql(self, sql_string: str) -> dict:
        items = self.__split_sql_string(sql_string)
        for clause in self.CLAUSES.keys():
            for idx, item in enumerate(items):
                if item.upper() == clause:
                    self.CLAUSES[item.upper()](items[idx + 1])
        return self.__to_dict()

    def __split_sql_string(self, sql_string: str) -> List[str]:
        sql_string = sql_string.replace(";", "")
        sql_string = sql_string.replace("\n", " ")
        _RE_COMBINE_WHITESPACE = re.compile(r"\s+")  # noqa
        sql_string = _RE_COMBINE_WHITESPACE.sub(" ", sql_string).strip()
        split_mask_clauses = create_mask(
            self.CLAUSES.keys(), can_start_sentence=True, keep_separators=True
        )
        return re.split(split_mask_clauses, sql_string)

    def __to_dict(self) -> dict:
        config = {
            "from": dict(self.__from),
            "select": dict(self.__select),
            "where": dict(self.__where),
            "join": {
                "inner": dict(self.__inner_join),
                "child": dict(self.__child_join),
                "parent": dict(self.__parent_join),
            },
        }
        config["join"] = {key: value for key, value in config["join"].items() if value}
        return {key: value for key, value in config.items() if value}

    def __select_parser(self, string):
        item_parsed = re.split(create_mask(",", optional_spaces=True), string)
        for item in item_parsed:
            alias, select_rule = re.split(r"\.", item, 1)
            assert (
                alias in self.__from.keys()
            ), f"Ressource {alias} was used in SELECT {item} but was not defined"
            self.__select[alias].append(f"{self.__from[alias]}.{select_rule}")

    def __from_parser(self, string):
        if "," not in string:
            table_aliases = [string]
        else:
            table_aliases = string.split(",")
        for table_alias in table_aliases:
            item_parsed = re.split(create_mask("AS"), table_alias)
            if len(item_parsed) == 2:
                resource_type = item_parsed[0]
                alias = item_parsed[1]
                self.__from[alias] = resource_type
            elif len(item_parsed) == 1:
                resource_type = item_parsed[0]
                alias = item_parsed[0]
                self.__from[alias] = resource_type
            else:
                raise ValueError(f"The FROM sub-clause {table_alias} couldn't be parsed properly")

    def __inner_join_parser(self, string):
        alias_parent, searchparam_parent, alias_child = self.__join_parser(string)
        self.__inner_join[alias_parent][searchparam_parent] = alias_child

    def __child_join_parser(self, string):
        alias_parent, searchparam_parent, alias_child = self.__join_parser(string)
        self.__child_join[alias_parent][searchparam_parent] = alias_child

    def __parent_join_parser(self, string):
        alias_parent, searchparam_parent, alias_child = self.__join_parser(string)
        self.__parent_join[alias_parent][searchparam_parent] = alias_child

    def __join_parser(self, string):
        item_parsed = re.split(create_mask("ON"), string)
        assert len(item_parsed) == 2, f"There was a problem with the JOIN statement: {string}"
        self.__from_parser(item_parsed[0])

        join_conditions = re.split(create_mask("=", optional_spaces=True), item_parsed[1])
        assert (
            len(join_conditions) == 2
        ), f"The JOIN condition {join_conditions} couldn't be parsed properly"

        alias_parent = None
        searchparam_parent = None
        alias_child = None
        for join_condition in join_conditions:
            alias, join_rule = re.split(r"\.", join_condition, 1)
            assert alias in self.__from.keys(), (
                f"The {alias} alias used in the {join_condition} join must have been referenced "
                "behind a FROM, INNER JOIN, PARENT JOIN or CHILD JOIN. "
                f"The only aliases referenced are here: {self.__from}"
            )
            if join_rule.strip() == "id":
                if alias_child:
                    raise ValueError
                alias_child = alias
            else:
                if alias_parent or searchparam_parent:
                    raise ValueError
                alias_parent = alias
                searchparam_parent = join_rule

        return alias_parent, searchparam_parent, alias_child

    def __where_parser(self, string):
        item_parsed = re.split(create_mask("AND"), string)
        for where_condition in item_parsed:
            condition_parsed = re.split(
                create_mask("=", keep_separators=True, optional_spaces=True), where_condition
            )
            if len(condition_parsed) != 3:
                condition_parsed = re.split(
                    create_mask(PREFIX, keep_separators=True), where_condition
                )
            assert (
                len(condition_parsed) == 3
            ), f"The WHERE condition {where_condition} couldn't be parsed properly"
            alias, where_rule = re.split(r"\.", condition_parsed[0], 1)
            quotes_match = re.search(r"^['\"](.*)['\"]$", condition_parsed[2])
            value = quotes_match.group(1) if quotes_match else condition_parsed[2]

            if where_rule not in self.__where[alias]:
                self.__where[alias][where_rule] = []

            if condition_parsed[1] in PREFIX:
                prefix = condition_parsed[1]
                value = {prefix: value}

            self.__where[alias][where_rule].append(value)

    def __order_by_parser(self, string):
        raise NotImplementedError("The ORDER BY keyword is not supported for the moment.")

    def __group_by_parser(self, string):
        raise NotImplementedError("The ORDER BY keyword is not supported for the moment.")

    def __union_parser(self, string):
        raise NotImplementedError("The UNION keyword is not supported for the moment.")

    def __limit_parser(self, string):
        raise NotImplementedError("The LIMIT keyword is not supported for the moment.")



"""classes representing the different types of information manipulated in FHIR2Dataset
"""
import re
import random
import logging
import networkx as nx
import matplotlib.pyplot as plt
from collections import defaultdict
from pprint import pformat
from dataclasses import dataclass, field, asdict
from typing import Type, List, Optional

from fhir2dataset.visualization_tools import custom_repr
from fhir2dataset.fhirpath import parse_fhirpath


logger = logging.getLogger(__name__)


@dataclass
class SearchParameter:
    code: Optional[str] = field(default=None)
    fhirpath: Optional[str] = field(default=None)
    resource_types: Optional[List[str]] = field(default=None)
    prefix: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)


@dataclass
class Element:
    """unit entity that can be found in a fhir instance in json format using fhirpath or in a table in the col_name column.

    An element is destined to be associated with a single value. For example, if we want to retrieve the patient id of a bundle containing 2 patients, we will have to create 2 element instances.
    """  # noqa

    col_name: str
    fhirpath: str
    goal: str = field(default="select")  # select, where or join
    value: Optional[list] = field(default=None)
    concat_type: Optional[str] = field(default="cell")
    search_parameter: Optional[SearchParameter] = field(default=None)


@dataclass
class Elements:
    """collection allowing to group together elements"""

    elements: List[Element] = field(default_factory=list)

    def append(self, x):
        self.elements.append(x)

    def get_subset_elements(self, goal):
        return [element for element in self.elements if element.goal == goal]

    def compute_forest_fhirpaths(self):
        forest = Forest()
        for fhirpath in [element.fhirpath for element in self.elements]:
            forest.add_fhirpath(fhirpath)
        forest.simplify_trees()
        forest.parse_fhirpaths()
        self.forest = forest
        self.forest_dict = forest.create_forest_dict()


@dataclass
class ResourceAliasInfoBasic:
    alias: str
    resource_type: str
    elements: Elements() = field(default_factory=Elements())


class ResourceAliasInfo(ResourceAliasInfoBasic):
    def __repr__(self):
        return custom_repr(super().__repr__())


@dataclass
class EdgeInfo:
    parent: str
    child: str
    searchparam_parent: Optional[str] = field(default=None)
    join_how: str = field(default="inner")
    searchparam_prefix: Optional[dict] = field(default=None)


class SearchParameters:
    def __init__(self, search_parameters: List[Type[SearchParameter]] = None):
        if search_parameters:
            self.items = search_parameters
        else:
            self.items = []
        self._init_data()

    def add(self, search_parameters):
        if isinstance(search_parameters, SearchParameter):
            self.items.append(search_parameters)
            self._add_data(search_parameters)
        elif isinstance(search_parameters, list):
            self.items.extend(search_parameters)
            for search_parameter in search_parameters:
                self._add_data(search_parameter)
        else:
            raise TypeError(
                f"{search_parameters} should be a list or a SearchParameter type instead of "
                f"{type(search_parameters)} type"
            )

    def searchparam_to_fhirpath(self, search_param: str, resource_type: str = "all"):
        """Retrieve the fhirpath that allows to retrieve the element that is the object of
        a searchparam in a json instance (after the 'resource' key) of a resource of a certain
        type

        Arguments:
            resource_type {str} -- name of a resource type (e.g. 'Organization')
            search_param {str} -- name of a searchparam of this resource type (e.g. 'address-postalcode')

        Returns:
            str -- the fhirpath for retrieving the element that is the subject of the searchparam
            (e.g. 'address.postalCode')
        """  # noqa
        try:
            return self._data[search_param][resource_type]
        except KeyError:
            logger.warning(f"The searchparam '{search_param}' doesn't exist in the rules")
            return None

    def _init_data(self):
        self._data = defaultdict(lambda: defaultdict(dict))
        for search_parameter in self.items:
            self._add_data(search_parameter)

    def _add_data(self, search_parameter: Type[SearchParameter]):
        fhirpath = search_parameter.fhirpath
        data = self._data[search_parameter.code]
        for resource_type in search_parameter.resource_types:
            if resource_type not in data.keys():
                data[resource_type] = fhirpath
            else:
                raise ValueError(
                    f"the search parameter {search_parameter.code} is already recorded\n"
                    f"data already recorded: {pformat(self._data[search_parameter.code])}\n"
                    f"data given as argument: {search_parameter}\n"
                )


@dataclass(eq=True, frozen=True)
class Node:
    """Modeling a node of a process tree"""

    fhirpath: str
    index: str
    previous_node_hash: int = 0


def break_parenthesis(exp: str):
    sublist = re.split(r"(\(|\))", exp)
    return sublist


def split_fhirpath(fhirpath: str) -> List[Node]:
    """Transforms a fhirpath into a list of sub-fhirpaths that run one after the other would give
    the same result as the initial fhirpath.

    Args:
        fhirpath (str): a string of characters representing a fhirpath

    Returns:
        List[Node]: a list of Nodes representing the sub-fhirpaths described above

    Examples:
        1. "Patient.name.given" becomes ["Patient","name","given"]
        1. "Patient.name | Practitioner.name" becomes ["Patient.name | Practitioner.name"]
        2. "Patient.(x | y).use" becomes ["Patient","(x | y)","use"]
    """
    sub_paths = [sub_path for sub_path in fhirpath.split(".") if sub_path != ""]

    tmps_list = []
    tmp_word = ""
    num_brackets = 0
    num_or = fhirpath.count("|")

    for sub_path in sub_paths:
        tmp_word = f"{tmp_word}.{sub_path}" if tmp_word else sub_path
        for character in sub_path:
            if character == "(":
                num_brackets += 1
            elif character == ")":
                num_brackets -= 1
            elif character == "|" and num_brackets % 2 == 1:
                num_or -= 1
        if not num_brackets:
            tmps_list.append(tmp_word)
            tmp_word = ""
    if num_or != 0:
        tmps_list = [fhirpath]

    node_list = _create_node_list(tmps_list)
    logger.debug(f"the fhirpath: {fhirpath} is parsed in {[node.fhirpath for node in node_list]}")
    return node_list


def _create_node_list(sub_fhirpaths: List[str]):
    node_list = []
    node = None
    for idx, fhirpath in enumerate(sub_fhirpaths):
        if node:
            node = Node(fhirpath=fhirpath, index=str(idx), previous_node_hash=hash(node))
        else:
            node = Node(fhirpath=fhirpath, index=str(idx))
        node_list.append(node)
    return node_list


class Forest:
    """class modeling the entire set of process trees

    Attributes:
        trees (dict): dictionary where the key is the value of the root and the value of the
        associated complete tree
        num_exp (int): total number of fhirpath that will be computed by this process tree forest
    """

    def __init__(self):
        self.trees = {}
        self.num_exp = 0

    def add_fhirpath(self, fhirpath: str) -> None:
        """Adding a fhirpath to the forest is done in the following steps:
            1. the fhirpath is divided into sub-fhirpaths which executed one after the other give
            the same result as the execution of the whole fhirpath.
            2. these subfhirpaths are added to the forest, each subfhirpath representing a node of
            a tree. The interest is that if 2 fhirpaths share the same first sub-fhirpaths then they
            will share the same first nodes of the tree.

        Args:
            fhirpath (str): string representing a fhirpath
        """
        splitted_fhirpath = split_fhirpath(fhirpath)
        self.__add_splitted_fhirpath(splitted_fhirpath)

    def simplify_trees(self):
        """This function, executed once all the fhirpaths have been added, allows to reduce the
        trees in the forest. If a node has only one successor node and if they are tagged with
        exactly the same column numbers (the same processes), then they will be merged.
        """
        new_trees = {}
        for root, tree in self.trees.items():
            tree.simplify_tree()
            new_trees[tree.root] = tree
        self.trees = new_trees

    def parse_fhirpaths(self):
        """transforms each expression that corresponds to a sub fhirpath of a node into a parsed
        version used by the fhirpath.js library.
        """
        for tree in self.trees.values():
            tree.parse_fhirpaths()

    def create_forest_dict(self):
        """creates a dictionary modeling the forest understandable by the javascript function in
        the forest.js file

        Returns:
            dict: dictionary described above
        """
        forest_dict = {}
        for root, tree in self.trees.items():
            nodes = tree.graph.nodes
            edges = tree.graph.edges

            nodes_dict = {}
            for node in nodes:
                nodes_dict[str(hash(node))] = asdict(node)
                nodes_dict[str(hash(node))]["column_idx"] = nodes[node]["column_idx"]
                nodes_dict[str(hash(node))]["parsed_fhirpath"] = nodes[node]["parsed_fhirpath"]

            edges_array = []
            for edge in edges:
                edges_array.append([str(hash(edge[0])), str(hash(edge[1]))])

            forest_dict[str(hash(root))] = {"nodes_dict": nodes_dict, "edges_array": edges_array}
        return forest_dict

    def __add_splitted_fhirpath(self, splitted_fhirpath=List[Node]):
        root = splitted_fhirpath[0]
        if root not in self.trees.keys():
            self.trees[root] = Tree(root)
        self.trees[root].add_edges_from_splitted_fhirpath(splitted_fhirpath, self.num_exp)
        self.num_exp += 1


class Tree:
    """class modeling a process tree

    Attributes:
        root (Node): the root of the tree
        graph (nx.DiGraph): oriented graph representing the tree
    """

    def __init__(self, root: Node):
        self.root = root
        self.graph = nx.DiGraph()

    def add_edges_from_splitted_fhirpath(self, splitted_fhirpath: list, column_idx: int):
        """add splitted_fhirpath to the tree, each sub-part of the fhirpath representing a node of
        the tree. The interest is that if 2 fhirpaths share the same first sub-fhirpaths then they
        will share the same first nodes of the tree

        Args:
            splitted_fhirpath (List[str]): list of fhirpaths representing a process
            column_idx ([type]): column number corresponding to the process number
        """
        for node in splitted_fhirpath:
            if node in self.graph.nodes:
                self.graph.nodes[node]["column_idx"].append(column_idx)
            else:
                self.graph.add_node(node, column_idx=[column_idx])
        edges = []
        for idx in range(len(splitted_fhirpath) - 1):
            edges.append((splitted_fhirpath[idx], splitted_fhirpath[idx + 1]))
        self.graph.add_edges_from(edges)

    def parse_fhirpaths(self):
        """transforms each expression that corresponds to a sub fhirpath of a node into a parsed
        version used by the fhirpath.js library.
        """
        for node in nx.dfs_preorder_nodes(self.graph, self.root):
            self.graph.nodes[node]["parsed_fhirpath"] = parse_fhirpath(node.fhirpath)

    def simplify_tree(self):
        """This function, executed once all the fhirpaths have been added, allows to reduce the
        tree. If a node has only one successor node and if they are tagged with exactly the same
        column numbers (the same processes), then they will be merged.
        """
        tmp_root = self.root
        final_root = self.root
        new_graph = self.graph.copy()

        previous_node_new_graph = None
        previous_node_created_new_graph = None
        column_idx = nx.get_node_attributes(self.graph, "column_idx")

        for node_old_graph in nx.dfs_preorder_nodes(self.graph, self.root):
            successors_old_graph = list(self.graph.successors(node_old_graph))
            if len(successors_old_graph) == 1:
                successor_old_graph = successors_old_graph[0]
                if column_idx[successor_old_graph] == column_idx[node_old_graph]:
                    if (
                        previous_node_created_new_graph
                        and column_idx[successor_old_graph] == column_idx[node_old_graph]
                    ):
                        previous_node_new_graph = previous_node_created_new_graph
                    else:
                        previous_node_new_graph = node_old_graph

                    new_fhirpath = (
                        f"{previous_node_new_graph.fhirpath}.{successor_old_graph.fhirpath}"
                    )
                    new_index = f"{previous_node_new_graph.index}-{successor_old_graph.index}"
                    new_node_new_graph = Node(new_fhirpath, new_index)

                    nx.relabel_nodes(
                        new_graph, {successor_old_graph: new_node_new_graph}, copy=False
                    )
                    new_graph = nx.contracted_nodes(
                        new_graph, new_node_new_graph, previous_node_new_graph, self_loops=False
                    )
                    previous_node_created_new_graph = new_node_new_graph
                    if node_old_graph == tmp_root:
                        tmp_root = successor_old_graph
                        final_root = new_node_new_graph
            else:
                previous_node_created_new_graph = None
        self.graph = new_graph
        self.root = final_root


def hierarchy_pos(G, root=None, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5):  # noqa

    """
    From Joel's answer at https://stackoverflow.com/a/29597209/2966723.
    Licensed under Creative Commons Attribution-Share Alike

    If the graph is a tree this will return the positions to plot this in a
    hierarchical layout.

    G: the graph (must be a tree)

    root: the root node of current branch
    - if the tree is directed and this is not given,
      the root will be found and used
    - if the tree is directed and this is given, then
      the positions will be just for the descendants of this node.
    - if the tree is undirected and not given,
      then a random choice will be used.

    width: horizontal space allocated for this branch - avoids overlap with other branches

    vert_gap: gap between levels of hierarchy

    vert_loc: vertical location of root

    xcenter: horizontal location of root
    """
    if not nx.is_tree(G):
        raise TypeError("cannot use hierarchy_pos on a graph that is not a tree")

    if root is None:
        if isinstance(G, nx.DiGraph):
            root = next(
                iter(nx.topological_sort(G))
            )  # allows back compatibility with nx version 1.11
        else:
            root = random.choice(list(G.nodes))

    def _hierarchy_pos(
        G, root, width=1.0, vert_gap=0.2, vert_loc=0, xcenter=0.5, pos=None, parent=None  # noqa
    ):
        """
        see hierarchy_pos docstring for most arguments

        pos: a dict saying where all nodes go if they have been assigned
        parent: parent of this branch. - only affects it if non-directed

        """

        if pos is None:
            pos = {root: (xcenter, vert_loc)}
        else:
            pos[root] = (xcenter, vert_loc)
        children = list(G.neighbors(root))
        if not isinstance(G, nx.DiGraph) and parent is not None:
            children.remove(parent)
        if len(children) != 0:
            dx = width / len(children)
            nextx = xcenter - width / 2 - dx / 2
            for child in children:
                nextx += dx
                pos = _hierarchy_pos(
                    G,
                    child,
                    width=dx,
                    vert_gap=vert_gap,
                    vert_loc=vert_loc - vert_gap,
                    xcenter=nextx,
                    pos=pos,
                    parent=root,
                )
        return pos

    return _hierarchy_pos(G, root, width, vert_gap, vert_loc, xcenter)


def show_tree(graph, number=None):
    plt.figure(figsize=(5, 5))
    layout = hierarchy_pos(graph)
    labels = nx.get_node_attributes(graph, "column_idx")
    for key, column_idx in labels.items():
        labels[key] = f"{key.fhirpath}\n{column_idx}"
    nx.draw_networkx(graph, labels=labels, pos=layout)
    plt.show()



import os
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
from setuptools.command.sdist import sdist

from subprocess import call

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"), encoding="utf-8"
) as f:
    long_description = f.read()


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class MyBuildCommand(build_py):
    def run(self):
        call(["npm", "install", "--prefix", "fhir2dataset/metadata"])
        build_py.run(self)


class MySdistCommand(sdist):
    def run(self):
        call(["npm", "install", "--prefix", "fhir2dataset/metadata"])
        sdist.run(self)


requirements = read("requirements.txt").split()

setup(
    cmdclass={"build_py": MyBuildCommand, "sdist": MySdistCommand},
    name="fhir2dataset",
    packages=find_packages(),
    include_package_data=True,
    version="0.1.5",
    license="Apache License 2.0",
    description="Transform FHIR to Dataset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Lucile Saulnier",
    author_email="contact@arkhn.com",
    url="https://github.com/arkhn/FHIR2Dataset",
    keywords=["arkhn", "medical", "fhir", "FHIR", "Dataset", "API"],
    install_requires=requirements,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
